import numpy as np
from scipy.interpolate import RectBivariateSpline
from .data_read import Metadata


class ARTOFTransformer:
    """Class to transform raw artof data"""

    def __init__(self, metadata: Metadata, x0: int = None, y0: int = None, t0: int = None):
        """Initializer ARTOFTransform class"""
        
        self.x_transform, self.y_transform, self.t_transform = self.ticks_to_SI_transform(metadata, x0, y0, t0)
                
        self.E_transform, self.theta_transform = self.tr_to_Etheta_transform(metadata)

    def transform(self, raw_data: list, load_as: str) -> list:
        """
        Transform raw data to desired representation.

        Args:
            raw_data: 2D list of raw data points (x, y, t ticks).
            load_as: Desired representation to transform to (options: 'raw', 'raw_SI', 'cylindrical', 'spherical').

        Returns:
            Three 2D list of transformed data, list of variable names, and list of default bin edges for given transformation.
        """
        match load_as:
            case 'raw':
                data = raw_data
            case 'raw_SI':
                x = self.x_transform.ev(raw_data[:,1], raw_data[:,0])
                y = self.y_transform.ev(raw_data[:,1], raw_data[:,0])
                t = self.t_transform(raw_data[:,2])

                data = np.stack([x, y, t], -1)
            case 'cylindrical':
                x = self.x_transform.ev(raw_data[:,1], raw_data[:,0])
                y = self.y_transform.ev(raw_data[:,1], raw_data[:,0])
                t = self.t_transform(raw_data[:,2])
                r, phi = self.xy_to_polar(x,y)

                data = np.stack([r, phi, t], -1)
            case 'spherical':
                x = self.x_transform.ev(raw_data[:,1], raw_data[:,0])
                y = self.y_transform.ev(raw_data[:,1], raw_data[:,0])
                t = self.t_transform(raw_data[:,2])
                r, phi = self.xy_to_polar(x,y)
                E = self.E_transform.ev(r, t)
                theta = self.theta_transform.ev(r, t)

                data = np.stack([E, phi, theta],-1)
            case _:
                print(f'Did not recognize transformation of type {load_as}. Using raw data')
                data = raw_data
        return data

    def get_axis_and_bins(self, load_as: str, metadata: Metadata) -> tuple:
        match load_as:
            case 'raw':
                return ['x_ticks', 'y_ticks', 't_ticks'], [[-1500, 1500, 101], [-1500, 1500, 101],[12000, 18000, 201]]
            case 'raw_SI':
                return ['x_m', 'y_m', 't_s'], [[-0.027, 0.027, 101], [-0.027, 0.027, 101],[0.3e-6, .4e-6, 201]]
            case 'cylindrical':
                return ['r_m', 'phi_rad', 't_s'], [[0, 0.027, 101], [-np.pi, np.pi, 201],[0.3e-6, .4e-6, 201]]
            case 'spherical':
                begin_energy, end_energy = metadata.general.spectrumBeginEnergy, metadata.general.spectrumEndEnergy
                theta_max = metadata.lensmode.maxTheta
                return ['E_eV', 'phi_rad', 'theta_rad'], [[begin_energy, end_energy, 101], [-np.pi, np.pi, 201],[0, theta_max, 201]] 

    def ticks_to_SI_transform(self, metadata: Metadata, x0: int = None, y0: int = None, t0: int = None) -> tuple:
        """
        Transform x, y, and t from ticks to SI units using transformation matrices and tdcResolution from acquisition.cfg file.

        Args:
            metadata: Metadata class containing all metadata for current measurement.
            x0: x offset in ticks (default: from the acquisition.cfg file).
            t0: y offset in ticks (default: from the acquisition.cfg file).
            t0: t offset in ticks (default: from the acquisition.cfg file).

        Returns:
            Three lists containing x, y, and t values in SI units.
        """
        # convert x and y ticks to radius in m and phi in radians
        detector = metadata.detector
        x0 = detector.x0 if x0 is None else x0
        y0 = detector.y0 if y0 is None else y0
        x_transform = self.create_matrix_transform(x0, y0, detector.transformXVector, detector.transformYVector, detector.transformXMatrix)
        y_transform = self.create_matrix_transform(x0, y0, detector.transformXVector, detector.transformYVector, detector.transformYMatrix)

        # transform time ticks to time in seconds
        t0 = detector.t0 if t0 is None else t0
        t_transform = lambda t:  self.transform_time(t, t0, detector.tdcResolution)

        return x_transform, y_transform, t_transform


    def xy_to_polar(self, x: float, y: float) -> tuple:    
        """
        Transform x and y in SI units to polar coordinates. The function arctan2(y, x) is used.

        Args:
            x: x value in meters (SI).
            y: y value in meters (SI).

        Returns:
            r in meters and phi in radians.
        """
        r = np.sqrt(x**2 + y**2)
        phi = np.arctan2(y, x)
        return r, phi


    def tr_to_Etheta_transform(self, metadata: Metadata) -> tuple:
        """
        Transform t and r in SI units to E and theta. The transformation matrices from the acquisition.cfg file are used.

        Args:
            metadata: Metadata class containing all metadata for current measurement.

        Returns:
            E in eV and theta in radians.
        """
        lensmode = metadata.lensmode
        # scale energy matrix and tof vector with energy scale centerEnergy/eKinRef
        energy_scale = metadata.general.centerEnergy/lensmode.eKinRef
        t_vector = lensmode.tofVector/np.sqrt(energy_scale)
        r_vector = lensmode.radiusVector
        energy_matrix = lensmode.energyMatrix*energy_scale
        theta_matrix = lensmode.thetaMatrix

        E = self.create_matrix_transform(0, 0, t_vector, r_vector, energy_matrix)
        theta = self.create_matrix_transform(0, 0, t_vector, r_vector, theta_matrix)
        return E, theta


    def transform_time(self, t_ticks: int, t0: int, tdcResolution: float) -> float:
        """
        Transform time from ticks to seconds.

        Args:
            t_ticks: Time in ticks.
            t0: Time offset in ticks.
            tdcResolution: Resolutions of time to digital converter (tdc); number of events per second.

        Returns:
            Time in seconds.
        """
        return (t_ticks - t0) * 1 / tdcResolution 


    def create_matrix_transform(self, p1_0: int, p2_0: int, p1_vec: list, p2_vec: list, trans_mat: list) -> RectBivariateSpline:
        """
        Transform 2D data point using a given matrix using interpolation through a bivariate spline.

        Args:
            p1_0: Offset of p1.
            p2_0: Offset of p2.
            p1_vec: Vector corresponding to p1 and the columns of the matrix.
            p2_vec: Vector corresponding to p2 and the rows of the matrix.
            trans_mat: 2D list representing the transformation matrix.

        Returns:
            RectBivariateSpline interpolation for given matrix.
        """
        interp = RectBivariateSpline(p2_vec-p2_0, p1_vec-p1_0, trans_mat)
        return interp


import numpy as np
import matplotlib.pyplot as plt
import threading

# internal
from .data_read import read_metadata, load_file
from .data_process import get_bin_edges, project_data
from .Transformer import ARTOFTransformer



# def process_data(dir, i, load_as, transformer):
#     raw_data = load_file(dir, i)
#     transformer.transform(raw_data, load_as)

class ARTOFLoader:
    """Class to load ARTOF data."""

    def __init__(self):
        """Initialize ARTOFLoader class"""
        self.binned_data = None
        self.bin_edges = list()
        self.axes = None
        self.metadata = None
        self.transformer = None

    def load_run(self, dir: str, load_as: str, x0: int = None, y0: int = None, t0: int = None, bin_confs: list = None):
        """
        Load ARTOF data for given run.

        Args:
            dir: Path to run directory.
            load_as: Load parameters in given format ('raw': x,y,t).
            bin_confs: List of 3 binning configurations for the 3 parameters (type defined by 'load_as').
        """
        # aquire metadata and configure transformer
        self.metadata = read_metadata(dir)
        self.transformer = ARTOFTransformer(self.metadata, x0, y0, t0)
        self.axes, def_bin_confs = self.transformer.get_axis_and_bins(load_as, self.metadata)
        bin_confs = def_bin_confs if bin_confs is None else bin_confs

        # transform data to desired format via multithreading
        data_pieces = []
        threads = []
        for i in range(self.metadata.general.lensIterations):
            thread = threading.Thread(target=self.__process_data, args=(dir, i, data_pieces, load_as))
            threads.append(thread)
            thread.start()
        for thread in threads:
            thread.join()
        data = np.concatenate(data_pieces, axis=0)

        # create bin edges based on the passed bin configs
        for i in range(3):
            self.bin_edges.append(get_bin_edges(data[:,i], bin_confs[i], data_id=self.axes[i]))
        # bin data in 3D histogram
        self.binned_data, _ = np.histogramdd(data, bins=self.bin_edges) 

    def __process_data(self, dir: str, i: int, data_pieces: list, load_as: str):        
        """
        Load and transform single data file in given format (needed for multithreading).

        Args:
            dir: Directory where data files are located.
            i: Index of the data to be loaded.
            data_pieces: List of transformed data pieces to which the newly transformed data should be appended.
            load_as: Desired transformation format.
        """
        raw_data = load_file(dir, i)
        data_pieces.append(self.transformer.transform(raw_data, load_as))



    def plot(self, axes: list, ranges: list = [None, None, None], width: float = 5.5, height: float = 5.5):
        """
        Plot loaded data as projection onto given axes. Projections are possible onto 1 or 2 axes.

        Args:
            axes: List containing all axes onto which the projection is performed, e.g., [0,1].
            ranges: List containing ranges for axes (e.g., [[50, 101], [0,50], None]), if None entire range of axes is used (default entire range of each axis).
            width: Width of plot (default 5).
            height: Height of plot (default 5).
        """
        
        proj_data = project_data(self.binned_data, axes, ranges)

        if len(axes) == 2: # plot data in 2D as image
            img_fig, img_ax = plt.subplots(figsize=(width, height))
            img_fig.subplots_adjust(bottom=0.2, left=0.2)
            img_ax.imshow(proj_data, cmap='terrain', norm='linear', origin='lower', extent=[self.bin_edges[axes[0]][0], self.bin_edges[axes[0]][-1], self.bin_edges[axes[1]][0], self.bin_edges[axes[1]][-1]], aspect="auto")
            img_ax.set_xlabel(self.__axis_label(self.axes[axes[0]]))
            img_ax.set_ylabel(self.__axis_label(self.axes[axes[1]]))
        elif len(axes) == 1: # plot data in 1D as line
            x_values = [(self.bin_edges[axes[0]][i] + self.bin_edges[axes[0]][i+1])/2 for i in range(len(self.bin_edges[axes[0]])-1)]            

            line_fig, line_ax = plt.subplots(figsize=(width, height))
            line_fig.subplots_adjust(bottom=0.2, left=0.2)
            line_ax.plot(x_values, proj_data)
            line_ax.set_xlabel(self.__axis_label(self.axes[axes[0]]))
            line_ax.set_ylabel('Counts')


        else:
            raise Exception(f'A projection along {len(axes)} axes is not possible.')
        
    def export_to_csv(self, path: str, axes: list, ranges: list = [None, None, None], delimiter: str = ','):
        """
        Export loaded data as projection onto given axes. Projections are possible onto 1 or 2 axes.

        Args:
            path: Path including file name to which the data is saved.
            axes: List containing all axes onto which the projection is performed, e.g., [0,1].
            ranges: List containing ranges for axes (e.g., [[50, 101], [0,50], None]), if None entire range of axes is used (default entire range of each axis).
            delimiter: Delimiter by which the data is separated (default ',').
        """  
        data_to_export = project_data(self.binned_data, axes, ranges)        
        np.savetxt(path, data_to_export, delimiter=delimiter)

    def __axis_label(self, axis: str) -> str:
        """
        Build string for matplotlib axis label including Greek characters.

        Args:
            axis: String containing the axis label and unit separated by '_'.

        Returns:
            Formatted string for matplotlib.
        """
        name, unit = axis.split('_')
        match name:
            case 'phi':
                name = '$\\varphi$'
            case 'theta':
                name = '$\\theta$'
        return f'{name} [{unit}]'






from dataclasses import dataclass, field
from datetime import datetime
import numpy as np

datetime_format = '%Y-%m-%d, %H:%M:%S'

# define metadata classes
@dataclass
class General:   
    version: str = ''
    acquisitionStarted: datetime = datetime(1, 1, 1, 0, 0, 0)
    spectrumBeginEnergy: float = 0.0
    spectrumEndEnergy: float = 0.0
    spectrumChannelWidthEnergy: float = 0.0
    lensLowEdgeEnergyStep: float = 0.0
    lensDwellType: str = ''
    lensDwellTime: int = 0
    lensIterations: int = 0
    lensSteps: int = 0
    userSpectrumEndEnergy: float = 0.0
    userLensLowEdgeEnergyStep: float = 0.0
    acquisitionMode: str = ''
    centerEnergy: int = 0
    xytFormat: str = ''
    conversionLibraryName: str = ''
    conversionLibraryVersion= str = ''

@dataclass
class Lensmode:
    lensK: float = 0.0
    vectorSize: int = 0
    maxTheta: float = 0.0
    eKinRef: int = 0
    tofVector: np.ndarray = field(default_factory= lambda: np.array([], dtype=float)) # 1D
    radiusVector: np.ndarray = field(default_factory= lambda: np.array([], dtype=float)) # 1D
    energyMatrix: np.ndarray = field(default_factory= lambda: np.array([], dtype=float)) # 2D
    thetaMatrix: np.ndarray = field(default_factory= lambda: np.array([], dtype=float)) # 2D

@dataclass
class Detector:
    transformVectorSize: int = 0
    transformMaxY: int = 0
    transformXRef: int = 0
    transformXVector: np.ndarray = field(default_factory= lambda: np.array([], dtype=float)) # 1D
    transformYVector: np.ndarray = field(default_factory= lambda: np.array([], dtype=float)) # 1D
    transformXMatrix: np.ndarray = field(default_factory= lambda: np.array([], dtype=float)) # 2D
    transformYMatrix: np.ndarray = field(default_factory= lambda: np.array([], dtype=float)) # 2D
    x0: int = 0
    y0: int = 0
    t0: int = 0
    t0Tolerance: int = 0
    tdcResolution: float = 0.0
    spatialResolution: int = 0
    spatialDiameter: int = 0
    xDelayTimeMin: int = 0
    xDelayTimeMax: int = 0
    yDelayTimeMin: int = 0
    yDelayTimeMax: int = 0
    zDelayTimeMin: int = 0
    zDelayTimeMax: int = 0

@dataclass
class Analyzer:
    elementSet: str = ''
    lensMode: str = ''
    passEnergy: int = 0  

@dataclass
class Metadata:
    """
    Metadata class containing all metadata for current measurement.
    """
    general: General = field(default_factory=lambda: General())
    lensmode: Lensmode = field(default_factory=lambda: Lensmode())
    detector: Detector = field(default_factory=lambda: Detector())
    analyzer: Analyzer = field(default_factory=lambda: Analyzer())

    
 
def read_metadata(dir: str) -> Metadata:
    """
    Read metadata for given run from acquisition.cfg file.

    Args:
        dir: Path to run directory.

    Returns:
        Metadata object containing all info from acquisition.cfg file.
    """
    # create empty metadata object
    metadata = Metadata()

    # read acquisition.cfg file
    with open(f'{dir}/acquisition.cfg') as f:
        cur_section = None
        while line := f.readline():
            # strip line of linebreaks
            line = line.rstrip()
            # set current section
            if line.startswith('['):
                cur_section = line[1:-1]
            else:             
                # retrieve parameter name and value
                par_name, value = line.split('=')         
                # set current dataclass depending on section
                cur_dataclass = None                          
                match cur_section:
                    case 'general':
                        cur_dataclass = metadata.general
                    case 'lensmode':
                        cur_dataclass = metadata.lensmode
                    case 'detector':
                        cur_dataclass = metadata.detector
                    case 'analyzer':
                        cur_dataclass = metadata.analyzer
                    case _:
                        print('Value either has no section or an unknown section. Skipping value.')
                        continue
                # set attribute in current dataclass
                setattr(cur_dataclass, par_name, parse_type(cur_dataclass, par_name, value))

    return metadata
        
def parse_type(dclass: 'dataclass', par_name: str, value: any) -> any:
    """
    Parse string read from file to given type. Possible non-trivial conversions are int, float, datetime, list.

    Args:
        dclass: Dataclass containing the given parameter.
        par_name: Name of parameter.
        value: Value to be set for parameter.

    Returns:
        Parsed parameter.
    """
    data_type = type(getattr(dclass, par_name)).__name__
    match data_type:
        case 'str':
            return value
        case 'int':
            return int(value)
        case 'float':
            return float(value)
        case 'datetime':
            return datetime.strptime(value, datetime_format)
        case 'ndarray':
            data = np.array(list(map(float, value.strip('[]').split(' '))))
            if par_name in ['energyMatrix', 'thetaMatrix']: # reorganize theta and enregy matrices from 1D list to 2D list
                data = np.reshape(data, (-1, dclass.vectorSize))
            elif par_name in ['transformXMatrix', 'transformYMatrix']: # reorganize transformation matrices from 1D list to 2D list
                data = np.reshape(data, (-1, dclass.transformVectorSize))
            return data
        case _:
            print(f'Did not find data type {data_type} for {par_name}, saving it as string.')
            return value
    

def load_file(dir: str, iteration: int) -> np.ndarray:
    """
    Load raw data from file and transform into data points with 3 values.

    Args:
        dir: Path to run directory.
        iteration: Current lens iteration.

    Returns:
        2D list containing three int32 values per row.
    """
    filepath = f'{dir}/{iteration}_0'
    raw_data = np.fromfile(filepath, dtype=np.int32)        
    # reshape long array into 2D array with 3 values per entry
    return np.reshape(raw_data, (-1, 3))


import numpy as np
    
def get_bin_edges(data: list, bin_config: list, data_id: str = "unknown") -> np.ndarray:
    """
    Create bin edges (limits) for given bin config.

    Args:
        data: 1D list of data values.
        bin_config: Bin config consisting of three values: [bin_min, bin_max, bin_number].
        data_id: Identifier to raise error message with identifier.

    Returns:
        List of bin edges.
    """
    bin_min, bin_max, bin_number = bin_config
    if type(bin_number) != int:
        raise Exception(f"The number of bins for {data_id} needs given as an int")
    # set min and max if not already given in bin_config
    if bin_min == None:
        bin_min = min(data)
    if bin_max == None:
        # ensuring that the highest values are not omitted
        bin_max = max(data)

    return np.linspace(bin_min, bin_max, bin_number+1)



    
def project_data(data: np.ndarray, axes: list, ranges: list) -> np.ndarray:
    """
    Plot loaded data in given projection. Projections are possible in 1 or 2 dimensions.

    Args:
        data: 3D data to be transformed.
        axes: List containing all axes onto which the projection is performed, e.g., [0,1].
        ranges: List of ranges for each axis.

    Returns:
        1 or 2D list containing the projection.
    """

    if data is None:
        raise Exception('Load the data before projecting the data.')         
    if len(axes) not in [1,2]:
        raise Exception(f'A projection onto {len(axes)} axes is not possible.')
    
    # determine ranges (None is entire range)
    start_1, end_1 = (0, data.shape[0]) if ranges[0] is None else (ranges[0][0], ranges[0][1])
    start_2, end_2 = (0, data.shape[1]) if ranges[1] is None else (ranges[1][0], ranges[1][1])
    start_3, end_3 = (0, data.shape[2]) if ranges[2] is None else (ranges[2][0], ranges[2][1])

    proj_data = data[start_1:end_1,start_2:end_2,start_3:end_3]
    
    if len(axes) == 2: # project data onto 2 axes
        # determine the projection axis
        proj_axis = 3 - sum(axes)
        # switch order of array if needed
        if axes[0] < axes[1]:
            proj_data = np.swapaxes(proj_data, axes[0], axes[1])
        proj_data = proj_data[:,:,:].sum(axis=proj_axis)
    elif len(axes) == 1: # project data onto 1 axis
        # get axes to project along
        proj_axes = [0,1,2]
        proj_axes.remove(axes[0])
        # project along axes
        proj_data = proj_data.sum(axis=tuple(proj_axes))


    return proj_data